@@ -5,7 +5,7 @@
  *                            | (__| |_| |  _ <| |___
  *                             \___|\___/|_| \_\_____|
  *
- * Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
@@ -33,6 +33,20 @@
 
 #ifdef _MSC_VER
 #define strncasecmp _strnicmp
+#define strcasecmp _stricmp
+#define strdup _strdup
+#endif
+
+#if CURL_AT_LEAST_VERSION(7,81,0)
+#define SUPPORTS_ZONEID
+#endif
+#if CURL_AT_LEAST_VERSION(7,80,0)
+#define SUPPORTS_URL_STRERROR
+#endif
+#if CURL_AT_LEAST_VERSION(7,78,0)
+#define SUPPORTS_ALLOW_SPACE
+#else
+#define CURLU_ALLOW_SPACE 0
 #endif
 
 #define OUTPUT_URL      0  /* default */
@@ -46,6 +60,7 @@
 #define OUTPUT_QUERY    8
 #define OUTPUT_FRAGMENT 9
 #define OUTPUT_ZONEID   10
+
 #define NUM_COMPONENTS 11 /* including "url" */
 
 #define PROGNAME        "trurl"
@@ -83,6 +98,17 @@ static const struct var variables[] = {
 #define ERROR_URL    7 /* could not get a URL out of the set components */
 #define ERROR_TRIM   8 /* a --trim problem */
 #define ERROR_BADURL 9 /* if --verify is set and the URL cannot parse */
+#define ERROR_GET   10 /* bad --get syntax */
+
+#ifndef SUPPORTS_URL_STRERROR
+/* provide a fake local mockup */
+static char *curl_url_strerror(CURLUcode error)
+{
+  static char buffer[128];
+  curl_msnprintf(buffer, sizeof(buffer), "URL error %u", (int)error);
+  return buffer;
+}
+#endif
 
 static void warnf(char *fmt, ...)
 {
@@ -110,22 +136,24 @@ static void help(void)
 {
   int i;
   fprintf(stderr, "Usage: " PROGNAME " [options] [URL]\n"
+          "  -a, --append [component]=[data] - append data to component\n"
           "  --accept-space               - give in to this URL abuse\n"
-          "  --append [component]=[data]  - append data to component\n"
           "  -f, --url-file [file/-]      - read URLs from file or stdin\n"
-          "  -g, --get [{component}s]     - output URL component(s)\n"
+          "  -g, --get [{component}s]     - output component(s)\n"
           "  -h, --help                   - this help\n"
-          " --json                        - output URL info as JSON\n"
-          "  --redirect [URL]             - redirect the base URL to this\n"
-          "  -s, --set [component]=[data] - set this component\n"
-          "  --trim [component]=[what]    - trim a component\n"
-          "  --url [base URL]             - URL to start with\n"
+          "  --json                       - output URL as JSON\n"
+          "  --query-separator [letter]   - if something else than '&'\n"
+          "  --redirect [URL]             - redirect to this\n"
+          "  -s, --set [component]=[data] - set component content\n"
+          "  --sort-query                 - alpha-sort the query pairs\n"
+          "  --trim [component]=[what]    - trim component\n"
+          "  --url [URL]                  - URL to work with\n"
           "  -v, --version                - show version\n"
-          " --verify                      - return error on (first) bad URL\n"
+          "  --verify                     - return error on (first) bad URL\n"
           " URL COMPONENTS:\n"
           "  "
     );
-  for(i=0; i< NUM_COMPONENTS; i++) {
+  for(i = 0; i< NUM_COMPONENTS; i++) {
     fprintf(stderr, "%s%s", i?", ":"", variables[i].name);
   }
   fprintf(stderr, "\n");
@@ -147,18 +175,26 @@ struct option {
   struct curl_slist *set_list;
   struct curl_slist *trim_list;
   const char *redirect;
+  const char *qsep;
   const char *format;
   FILE *url;
   bool urlopen;
   bool jsonout;
   bool verify;
   bool accept_space;
+  bool sort_query;
+  bool end_of_options;
   unsigned char output;
 
   /* -- stats -- */
   unsigned int urls;
 };
 
+#define MAX_QPAIRS 1000
+char *qpairs[MAX_QPAIRS]; /* encoded */
+char *qpairsdec[MAX_QPAIRS]; /* decoded */
+int nqpairs; /* how many is stored */
+
 static void urladd(struct option *o, const char *url)
 {
   struct curl_slist *n;
@@ -194,7 +230,7 @@ static void pathadd(struct option *o, const char *path)
     if(n) {
       o->append_path = n;
     }
-    free(urle);
+    curl_free(urle);
   }
 }
 
@@ -218,16 +254,16 @@ static void queryadd(struct option *o, const char *query)
     if(n) {
       o->append_query = n;
     }
-    free(urle);
+    curl_free(urle);
   }
 }
 
 static void appendadd(struct option *o,
                       const char *arg)
 {
-  if(!strncasecmp("path=", arg, 5))
+  if(!strncmp("path=", arg, 5))
     pathadd(o, arg + 5);
-  else if(!strncasecmp("query=", arg, 6))
+  else if(!strncmp("query=", arg, 6))
     queryadd(o, arg + 6);
   else
     errorf(ERROR_APPEND, "--append unsupported component: %s", arg);
@@ -270,7 +306,9 @@ static int getarg(struct option *op,
 {
   *usedarg = false;
 
-  if(!strcmp("-v", flag) || !strcmp("--version", flag))
+  if(!strcmp("--", flag))
+    op->end_of_options = true;
+  else if(!strcmp("-v", flag) || !strcmp("--version", flag))
     show_version();
   else if(!strcmp("-h", flag) || !strcmp("--help", flag))
     help();
@@ -283,7 +321,8 @@ static int getarg(struct option *op,
     urlfile(op, arg);
     *usedarg = 1;
   }
-  else if(checkoptarg("--append", flag, arg)) {
+  else if(checkoptarg("-a", flag, arg) ||
+          checkoptarg("--append", flag, arg)) {
     appendadd(op, arg);
     *usedarg = 1;
   }
@@ -298,6 +337,14 @@ static int getarg(struct option *op,
     op->redirect = arg;
     *usedarg = 1;
   }
+  else if(checkoptarg("--query-separator", flag, arg)) {
+    if(op->qsep)
+      errorf(ERROR_FLAG, "only one --query-separator is supported");
+    if(strlen(arg) != 1)
+      errorf(ERROR_FLAG, "only single-letter query separators are supported");
+    op->qsep = arg;
+    *usedarg = 1;
+  }
   else if(checkoptarg("--trim", flag, arg)) {
     trimadd(op, arg);
     *usedarg = 1;
@@ -306,20 +353,53 @@ static int getarg(struct option *op,
           checkoptarg("--get", flag, arg)) {
     if(op->format)
       errorf(ERROR_FLAG, "only one --get is supported");
+    if(op->jsonout)
+      errorf(ERROR_FLAG, "--get is mututally exclusive with --json");
     op->format = arg;
     *usedarg = 1;
   }
-  else if(!strcmp("--json", flag))
+  else if(!strcmp("--json", flag)) {
+    if(op->format)
+      errorf(ERROR_FLAG, "--json is mututally exclusive with --get");
     op->jsonout = true;
+  }
   else if(!strcmp("--verify", flag))
     op->verify = true;
-  else if(!strcmp("--accept-space", flag))
+  else if(!strcmp("--accept-space", flag)) {
+#ifdef SUPPORTS_ALLOW_SPACE
     op->accept_space = true;
+#else
+    warnf("built with too old libcurl version, --accept-space does not work");
+#endif
+  }
+  else if(!strcmp("--sort-query", flag))
+    op->sort_query = true;
   else
     return 1;  /* unrecognized option */
   return 0;
 }
 
+static void showqkey(const char *key, size_t klen, bool urldecode)
+{
+  int i;
+  for(i = 0; i< nqpairs; i++) {
+    if(urldecode) {
+      if(!strncmp(key, qpairsdec[i], klen) &&
+         (qpairsdec[i][klen] == '=')) {
+        fputs(&qpairsdec[i][klen + 1], stdout);
+        break;
+      }
+    }
+    else {
+      if(!strncmp(key, qpairs[i], klen) &&
+         (qpairs[i][klen] == '=')) {
+        fputs(&qpairs[i][klen + 1], stdout);
+        break;
+      }
+    }
+  }
+}
+
 static void get(struct option *op, CURLU *uh)
 {
   FILE *stream = stdout;
@@ -336,6 +416,7 @@ static void get(struct option *op, CURLU *uh)
       else {
         /* this is meant as a variable to output */
         char *end;
+        char *cl;
         size_t vlen;
         int i;
         bool urldecode = true;
@@ -351,34 +432,47 @@ static void get(struct option *op, CURLU *uh)
           ptr++;
         }
         vlen = end - ptr;
-        for(i = 0; variables[i].name; i++) {
-          if((strlen(variables[i].name) == vlen) &&
-             !strncasecmp(ptr, variables[i].name, vlen)) {
-            char *nurl;
-            CURLUcode rc;
-            rc = curl_url_get(uh, variables[i].part, &nurl,
-                              CURLU_DEFAULT_PORT|
-                              CURLU_NO_DEFAULT_PORT|
-                              (urldecode?CURLU_URLDECODE:0));
-            switch(rc) {
-            case CURLUE_OK:
-              fprintf(stream, "%s", nurl);
-              curl_free(nurl);
-            case CURLUE_NO_SCHEME:
-            case CURLUE_NO_USER:
-            case CURLUE_NO_PASSWORD:
-            case CURLUE_NO_OPTIONS:
-            case CURLUE_NO_HOST:
-            case CURLUE_NO_PORT:
-            case CURLUE_NO_QUERY:
-            case CURLUE_NO_FRAGMENT:
-            case CURLUE_NO_ZONEID:
-              /* silently ignore */
-              break;
-            default:
-              fprintf(stderr, PROGNAME ": %s (%s)\n", curl_url_strerror(rc),
-                      variables[i].name);
-              break;
+        /* check for a colon within here */
+        cl = memchr(ptr, ':', vlen);
+        if(cl) {
+          /* deduct the colon part */
+          if(!strncmp(ptr, "query:", 6))
+            showqkey(&ptr[6], end - cl - 1, urldecode);
+          else
+            errorf(ERROR_GET, "Bad --get syntax: %s", ptr);
+        }
+        else {
+          for(i = 0; variables[i].name; i++) {
+            if((strlen(variables[i].name) == vlen) &&
+               !strncmp(ptr, variables[i].name, vlen)) {
+              char *nurl;
+              CURLUcode rc;
+              rc = curl_url_get(uh, variables[i].part, &nurl,
+                                CURLU_DEFAULT_PORT|
+                                CURLU_NO_DEFAULT_PORT|
+                                (urldecode?CURLU_URLDECODE:0));
+              switch(rc) {
+              case CURLUE_OK:
+                fprintf(stream, "%s", nurl);
+                curl_free(nurl);
+              case CURLUE_NO_SCHEME:
+              case CURLUE_NO_USER:
+              case CURLUE_NO_PASSWORD:
+              case CURLUE_NO_OPTIONS:
+              case CURLUE_NO_HOST:
+              case CURLUE_NO_PORT:
+              case CURLUE_NO_QUERY:
+              case CURLUE_NO_FRAGMENT:
+#ifdef SUPPORTS_ZONEID
+              case CURLUE_NO_ZONEID:
+#endif
+                /* silently ignore */
+                break;
+              default:
+                fprintf(stderr, PROGNAME ": %s (%s)\n", curl_url_strerror(rc),
+                        variables[i].name);
+                break;
+              }
             }
           }
         }
@@ -418,7 +512,7 @@ static void set(CURLU *uh,
   struct curl_slist *node;
   bool varset[NUM_COMPONENTS];
   memset(varset, 0, sizeof(varset));
-  for(node =  o->set_list; node; node=node->next) {
+  for(node =  o->set_list; node; node = node->next) {
     char *set = node->data;
     int i;
     char *ptr = strchr(set, '=');
@@ -430,15 +524,19 @@ static void set(CURLU *uh,
         urlencode = false;
         vlen--;
       }
-      for(i=0; variables[i].name; i++) {
+      for(i = 0; variables[i].name; i++) {
         if((strlen(variables[i].name) == vlen) &&
-           !strncasecmp(set, variables[i].name, vlen)) {
+           !strncmp(set, variables[i].name, vlen)) {
+          CURLUcode rc;
           if(varset[i])
             errorf(ERROR_SET, "A component can only be set once per URL (%s)",
                    variables[i].name);
-          curl_url_set(uh, variables[i].part, ptr[1] ? &ptr[1] : NULL,
-                       CURLU_NON_SUPPORT_SCHEME|
-                       (urlencode ? CURLU_URLENCODE : 0) );
+          rc = curl_url_set(uh, variables[i].part, ptr[1] ? &ptr[1] : NULL,
+                            CURLU_NON_SUPPORT_SCHEME|
+                            (urlencode ? CURLU_URLENCODE : 0) );
+          if(rc)
+            warnf("Error setting %s: %s", variables[i].name,
+                  curl_url_strerror(rc));
           found = true;
           varset[i] = true;
           break;
@@ -452,13 +550,19 @@ static void set(CURLU *uh,
   }
 }
 
-static void jsonString(FILE *stream, const char *in, bool lowercase)
+static void jsonString(FILE *stream, const char *in, size_t len,
+                       bool lowercase)
 {
-  const char *i = in;
-  const char *in_end = in + strlen(in);
+  const unsigned char *i = (unsigned char *)in;
+  const char *in_end;
+
+  if(!len)
+    in_end = in + strlen(in);
+  else
+    in_end = &in[len];
 
   fputc('\"', stream);
-  for(; i < in_end; i++) {
+  for(; i < (unsigned char *)in_end; i++) {
     switch(*i) {
     case '\\':
       fputs("\\\\", stream);
@@ -482,9 +586,8 @@ static void jsonString(FILE *stream, const char *in, bool lowercase)
       fputs("\\t", stream);
       break;
     default:
-      if (*i < 32) {
+      if (*i < 32)
         fprintf(stream, "u%04x", *i);
-      }
       else {
         char out = *i;
         if(lowercase && (out >= 'A' && out <= 'Z'))
@@ -512,19 +615,36 @@ static void json(struct option *o, CURLU *uh)
       if(i)
         fputs(",\n", stdout);
       printf("    \"%s\": ", variables[i].name);
-      jsonString(stdout, nurl, false);
+      jsonString(stdout, nurl, 0, false);
     }
   }
+  if(nqpairs) {
+    int i;
+    fputs(",\n    \"params\": [\n", stdout);
+    for(i = 0 ; i < nqpairs; i++) {
+      char *sep = strchr(qpairsdec[i], '=');
+      if(i)
+        fputs(",\n", stdout);
+      fputs("      {\n        \"key\": ", stdout);
+      jsonString(stdout, qpairsdec[i],
+                 sep ? (size_t)(sep - qpairsdec[i]) : strlen(qpairsdec[i]),
+                 false);
+      fputs(",\n        \"value\": ", stdout);
+      jsonString(stdout, sep ? sep + 1 : "", 0, false);
+      fputs("\n      }", stdout);
+    }
+    fputs("\n    ]", stdout);
+  }
   fputs("\n  }", stdout);
 }
 
 /* --trim query="utm_*" */
-static void trim(CURLU *uh, struct option *o)
+static void trim(struct option *o)
 {
   struct curl_slist *node;
-  for(node = o->trim_list; node; node=node->next) {
+  for(node = o->trim_list; node; node = node->next) {
     char *instr = node->data;
-    if(strncasecmp(instr, "query", 5))
+    if(strncmp(instr, "query", 5))
       /* for now we can only trim query components */
       errorf(ERROR_TRIM, "Unsupported trim component: %s", instr);
     char *ptr = strchr(instr, '=');
@@ -533,64 +653,171 @@ static void trim(CURLU *uh, struct option *o)
          asterisk */
       size_t inslen;
       bool pattern;
-      char *q;
-      CURLUcode rc;
+      int i;
 
       ptr++; /* pass the = */
       inslen = strlen(ptr);
       pattern = ptr[inslen - 1] == '*';
       if(pattern)
         inslen--;
-      rc = curl_url_get(uh, CURLUPART_QUERY, &q, 0);
-      if(!rc && q) {
-        bool updated = 0;
-        char *newq = q;
 
-        do {
-          char *end = strchr(q, '&');
-          char *sep = strchr(q, '=');
+      for(i = 0 ; i < nqpairs; i++) {
+        char *q = qpairs[i];
+        char *sep = strchr(q, '=');
+        size_t qlen;
+        if(sep)
+          qlen = sep - q;
+        else
+          qlen = strlen(q);
 
-          if (! end) {
-            end = q + strlen(q);
-          }
-          if (! sep || sep > end) {
-            sep = end;
-          }
-
-          size_t qlen = sep - q;
-          if((pattern && (inslen <= qlen) &&
-              !strncasecmp(q, ptr, inslen)) ||
-              (!pattern && (inslen == qlen) &&
-              !strncasecmp(q, ptr, inslen))) {
-              /* this part should be stripped out */
-              if(end && *end) {
-                end++;
-                /* move the rest of the string to this position */
-                memmove(q, end, strlen(end)+1);
-              }
-              else {
-                /* don't leave a trailing ampersand */
-                if((q > newq) && (q[-1] == '&'))
-                  q--;
-                *q = 0;
-              }
-              updated = true;
-          }
-          else {
-            q = end;
-            if(!q)
-              break;
-            q++;
-          }
-        } while(*q);
-        if(updated)
-          rc = curl_url_set(uh, CURLUPART_QUERY, newq, 0);
-        curl_free(newq);
+        if((pattern && (inslen <= qlen) &&
+            !strncasecmp(q, ptr, inslen)) ||
+           (!pattern && (inslen == qlen) &&
+            !strncasecmp(q, ptr, inslen))) {
+          /* this qpair should be stripped out */
+          free(qpairs[i]);
+          free(qpairsdec[i]);
+          qpairs[i] = strdup(""); /* marked as deleted */
+          qpairsdec[i] = strdup(""); /* marked as deleted */
+        }
       }
     }
   }
 }
 
+/* memdup the amount and add a trailing zero */
+static char *memdupzero(char *source, size_t len)
+{
+  char *p = malloc(len + 1);
+  if(p) {
+    memcpy(p, source, len);
+    p[len] = 0;
+    return p;
+  }
+  return NULL;
+}
+
+/* URL decode the pair and return it in an allocated chunk */
+static char *memdupdec(char *source, size_t len)
+{
+  char *sep = memchr(source, '=', len);
+  char *left = NULL;
+  char *right = NULL;
+  char *str, *dup;
+  int leftlen = 0;
+  int rightlen = 0;
+
+  left = curl_easy_unescape(NULL, source, sep ? (size_t)(sep - source) : len,
+                            &leftlen);
+  if(sep)
+    right = curl_easy_unescape(NULL, sep + 1, len - (sep - source) - 1,
+                               &rightlen);
+
+  str = curl_maprintf("%.*s%s%.*s", leftlen, left,
+                      right ? "=":"",
+                      rightlen, right?right:"");
+  curl_free(right);
+  curl_free(left);
+  dup = strdup(str);
+  curl_free(str);
+  return dup;
+}
+
+
+static void freeqpairs(void)
+{
+  int i;
+  for(i = 0; i<nqpairs; i++) {
+    free(qpairs[i]);
+    qpairs[i] = NULL;
+    free(qpairsdec[i]);
+    qpairsdec[i] = NULL;
+  }
+  nqpairs = 0;
+}
+
+/* store the pair both encoded and decoded */
+static char *addqpair(char *pair, size_t len)
+{
+  char *p = NULL;
+  char *pdec = NULL;
+  if(nqpairs < MAX_QPAIRS) {
+    p = memdupzero(pair, len);
+    pdec = memdupdec(pair, len);
+    if(p && pdec) {
+      qpairs[nqpairs] = p;
+      qpairsdec[nqpairs] = pdec;
+      nqpairs++;
+    }
+  }
+  else
+    warnf("too many query pairs");
+  return p;
+}
+
+/* convert the query string into an array of name=data pair */
+static void extractqpairs(CURLU *uh, struct option *o)
+{
+  char *q = NULL;
+  memset(qpairs, 0, sizeof(qpairs));
+  nqpairs = 0;
+  /* extract the query */
+  if(!curl_url_get(uh, CURLUPART_QUERY, &q, 0)) {
+    char *p = q;
+    char *amp;
+    while(*p) {
+      size_t len;
+      amp = strchr(p, o->qsep[0]);
+      if(!amp)
+        len = strlen(p);
+      else
+        len = amp - p;
+      addqpair(p, len);
+      if(amp)
+        p = amp + 1;
+      else
+        break;
+    }
+  }
+  curl_free(q);
+}
+
+static void qpair2query(CURLU *uh, struct option *o)
+{
+  int i;
+  int rc;
+  char *nq = NULL;
+  char *oldnq;
+  for(i = 0; i<nqpairs; i++) {
+    oldnq = nq;
+    nq = curl_maprintf("%s%s%s", nq?nq:"",
+                       (nq && *nq && *qpairs[i])? o->qsep: "", qpairs[i]);
+    curl_free(oldnq);
+  }
+  if(nq) {
+    rc = curl_url_set(uh, CURLUPART_QUERY, nq, 0);
+    if(rc)
+      warnf("internal problem");
+  }
+  curl_free(nq);
+}
+
+/* sort case insensitively */
+static int cmpfunc(const void *p1, const void *p2)
+{
+  return strcasecmp(*(const char **)p1,
+                    *(const char **)p2);
+}
+
+static void sortquery(struct option *o)
+{
+  if(o->sort_query) {
+    /* not these two lists may no longer be the same order after the sort */
+    qsort(&qpairs[0], nqpairs, sizeof(char *), cmpfunc);
+    qsort(&qpairsdec[0], nqpairs, sizeof(char *), cmpfunc);
+  }
+}
+
 static void singleurl(struct option *o,
                       const char *url) /* might be NULL */
 {
@@ -608,6 +835,7 @@ static void singleurl(struct option *o,
         if(o->verify)
           errorf(ERROR_BADURL, "%s [%s]", curl_url_strerror(rc), url);
         warnf("%s [%s]", curl_url_strerror(rc), url);
+        return;
       }
       else {
         if(o->redirect)
@@ -619,7 +847,7 @@ static void singleurl(struct option *o,
     set(uh, o);
 
     /* append path segments */
-    for(p = o->append_path; p; p=p->next) {
+    for(p = o->append_path; p; p = p->next) {
       char *apath = p->data;
       char *opath;
       char *npath;
@@ -643,29 +871,20 @@ static void singleurl(struct option *o,
       curl_free(opath);
     }
 
+    extractqpairs(uh, o);
+
     /* append query segments */
-    for(p = o->append_query; p; p=p->next) {
-      char *aq = p->data;
-      char *oq;
-      char *nq;
-      /* extract the current query */
-      if(!curl_url_get(uh, CURLUPART_QUERY, &oq, 0)) {
-        /* append the new segment */
-        nq = curl_maprintf("%s&%s", oq, aq);
-        if(nq)
-          /* set the new query */
-          curl_url_set(uh, CURLUPART_QUERY, nq, 0);
-        curl_free(nq);
-        curl_free(oq);
-      }
-      else {
-        /* no existing query, set the new one */
-        curl_url_set(uh, CURLUPART_QUERY, aq, 0);
-      }
+    for(p = o->append_query; p; p = p->next) {
+      addqpair(p->data, strlen(p->data));
     }
 
     /* trim parts */
-    trim(uh, o);
+    trim(o);
+
+    sortquery(o);
+
+    /* put the query back */
+    qpair2query(uh, o);
 
     if(o->jsonout)
       json(o, uh);
@@ -684,6 +903,11 @@ static void singleurl(struct option *o,
         errorf(ERROR_URL, "not enough input for a URL");
       }
     }
+
+    fflush(stdout);
+
+    freeqpairs();
+
     o->urls++;
     curl_url_cleanup(uh);
 }
@@ -698,9 +922,7 @@ int main(int argc, const char **argv)
 
   for(argc--, argv++; argc > 0; argc--, argv++) {
     bool usedarg = false;
-    if((argv[0][0] == '-' && argv[0][1] != '-') ||
-       /* single-dash prefix */
-       (argv[0][0] == '-' && argv[0][1] == '-')) {
+    if(!o.end_of_options && argv[0][0] == '-') {
       /* dash-dash prefixed */
       if(getarg(&o, argv[0], argv[1], &usedarg))
         errorf(ERROR_FLAG, "unknown option: %s", argv[0]);
@@ -715,6 +937,8 @@ int main(int argc, const char **argv)
       argv++;
     }
   }
+  if(!o.qsep)
+    o.qsep = "&";
 
   if(o.jsonout)
     fputs("[\n", stdout);
